const std = @import("std");
const mem = std.mem;
const heap = std.heap;

pub const c = @cImport({
    @cInclude("libvirt/libvirt.h");
    @cInclude("libvirt/libvirt-admin.h");
    @cInclude("libvirt/libvirt-lxc.h");
    @cInclude("libvirt/libvirt-qemu.h");
    @cInclude("libvirt/virterror.h");
});

fn handleError() VirError {
    const err = c.virGetLastError();
    std.debug.print("err: {any}\n", .{err});
    return VirError.OK;
    // switch (err.*.code) {
    //     0 => return VirError.OK,
    //     1 => return VirError.InternalError,
    //     2 => return VirError.NoMemory,
    //     // TODO rest
    //     else => VirError.OK,
    // }
}

pub const VirError = error{
    OK,
    InternalError,
    NoMemory,
    NoSupport,
    UnknownHost,
    NoConnect,
    InvalidConn,
    InvalidDomain,
    InvalidArg,
    OperationFailed,
    GetFailed,
    PostFailed,
    HttpError,
    SExprError,
    NoXen,
    XenCall,
    OsType,
    NoKernel,
    NoRoot,
    NoSource,
    NoTarget,
    NoName,
    NoOs,
    NoDevice,
    NoXenstore,
    DriverFull,
    CallFailed,
    XmlError,
    DomExist,
    OperationDenied,
    OpenFailed,
    ReadFailed,
    ParseFailed,
    ConfSyntax,
    // TODO rest
};

pub const Connection = struct {
    conn: c.virConnectPtr,
    allocator: mem.Allocator,

    pub fn connect(uri: []const u8, allocator: mem.Allocator) VirError!Connection {
        const connection = c.virConnectOpenAuth(@ptrCast(uri), c.virConnectAuthPtrDefault, 0);
        if (connection) |conn| return .{
            .conn = conn,
            .allocator = allocator,
        } else return handleError();
    }

    pub fn close(self: *const Connection) void {
        _ = c.virConnectClose(self.conn);
    }

    pub fn getURI(self: *const Connection) ![]u8 {
        const uri = c.virConnectGetURI(self.conn);
        defer std.c.free(uri);
        return try self.allocator.dupe(u8, mem.span(uri));
    }

    pub fn freeURI(self: *const Connection, uri: []u8) void {
        self.allocator.free(uri);
    }

    pub fn numOfDomains(self: *const Connection) !u32 {
        return @intCast(c.virConnectNumOfDomains(self.conn));
    }

    pub fn numOfDefinedDomains(self: *const Connection) !u32 {
        return @intCast(c.virConnectNumOfDefinedDomains(self.conn));
    }

    pub fn iterateDomains(self: *const Connection, flags: []const Domain.Flags) Domain.Iterator {
        var list: [*]c.virDomainPtr = undefined;
        var flags_int: c_uint = 0;
        for (flags) |f| flags_int |= @intFromEnum(f);
        const num = c.virConnectListAllDomains(self.conn, @ptrCast(&list), flags_int);
        return .{
            .list = list,
            .num = num,
            .curr = 0,
        };
    }
};

pub const Domain = struct {
    ptr: c.virDomainPtr,

    pub fn getName(self: *const Domain) []const u8 {
        const name = c.virDomainGetName(self.ptr);
        return mem.span(name);
    }

    pub fn isActive(self: *const Domain) bool {
        const active = c.virDomainIsActive(self.ptr);
        return if (active == 0) false else true;
    }

    pub const Flags = enum(c_uint) {
        ListDomainsActive = c.VIR_CONNECT_LIST_DOMAINS_ACTIVE,
        ListDomainsInactive = c.VIR_CONNECT_LIST_DOMAINS_INACTIVE,
    };

    pub const Iterator = struct {
        list: [*]c.virDomainPtr,
        num: c_int,
        curr: usize,

        pub fn deinit(self: *Iterator) void {
            var i: usize = 0;
            while (i < self.num) : (i += 1) _ = c.virDomainFree(self.list[i]);
        }

        pub fn first(self: *Iterator) Domain {
            self.curr = 0;
            return .{ .ptr = self.list[self.curr] };
        }

        pub fn next(self: *Iterator) ?Domain {
            if (self.curr >= self.num) return null;
            const ptr = self.list[self.curr];
            self.curr += 1;
            return .{ .ptr = ptr };
        }
    };
};
